/* //////////////////////////////////////////////////////////////////////////////////////
** File:      Dino.c
*/                                  
#define  Ver "Dino 18 ver 2.51\r       by Guiott"
/* Autore:    Guido Ottaviani-->g.ottaviani@mediaprogetti.it<--
** Descrizione:Gestione Robot Dino, comportamento tipo explorer   
**             questo e' un laboratorio per la prova di utilizzo di diversi sensori
**             e il movimento in un abiente sconosciuto
**              
**
** LA DESCRIZIONE PROSEGUE SU "settings.h"
///////////////////////////////////////////////////////////////////////////////////////*/


// include standard
#include <p18f4525.h>
#include <stdlib.h>
// #include <math.h>
#include <pwm.h>
#include <i2c.h>

#include "settings.h" // contiene: definizioni, prototipi funzioni, settings, descrizioni
#include "terminal.h" // contiene routine gestione tastiera e display
#include "myi2c.h"    // contiene routine gestione i2c a basso ed alto livello
#include "sensor.h"   // contiene routine gestione sensori peer explorer
#include "movement.h" // contiene routine gestione movimento

/* MAIN =============================================================================*/
void main (void)
{	
INTCONbits.GIEH=0;	// disabilitati tutti gli interrupt high
INTCONbits.GIEL=0;	// disabilitati tutti gli interrupt low	


Settings(); // imposta porte e registri (settings.h)


// ---- Inizializzazione flag e variabili
MotEnable = 0;			// motori disabilitati
PidTick=0;

SetDCPWM2(0x0100);   // motori fermi
SetDCPWM1(0x0100);

Buzzer=0;				// beeper spento
FlagTargetLight = 1; // i sensori di luce sono abilitati
LedRossoOFF;			// led spenti
LedGialloOFF;
LedVerdeOFF; 

// ---- Lettura costanti da EEPROM
Light1=ReadEeprom(0);// prima soglia di intensita' luminosa, il bot si gira verso la luce (37-0x25)
Light2=ReadEeprom(1);// seconda soglia, il bot ha raggiunto il target e si ferma a Xcm (48-0x30)
GasOn=ReadEeprom(2); // soglia di "gas rivelato" (130 - 0x82)
kp=ReadEeprom(3);    // costante errore proporzionale (fattore P) moltiplicato 16 (40 - 0x28)
ki=ReadEeprom(4);    // costante errore integrale (fattore I) moltiplicato 16 (24 - 0x18)
kd=ReadEeprom(5);    // costante errore derivativo (fattore D) moltiplicato 16 (24 - 0x18)
En=ReadEeprom(6);    // abilitazione motori, = 0 per debug o misure

// ---- init I2C
I2cBusCollFlag = 0;	
I2cEventFlag = 0;
I2cBusyFlag = 0;
for (i=0; i<I2cMaxDev; i++)
{
	I2c[i].Flag.Tx = 0;
	I2c[i].Flag.Rx = 0;
	for (j=0; j<=1; j++)
	{
		I2c[i].RxBuff[j] = 0;
	}
	
	for (j=0; j<=3; j++)
	{
		I2c[i].TxBuff[j] = 0;
	}
}


// ---- init LCD
InitFlag=1;			   // avvia l'inizializzazione dell'LCD 
LcdInitStatus=0;	   // azzera contatore stati init LCD
DisplayStatus=1;     // abilita la gestione del display all'avvio
UserInterfaceFlag=1; // abilita la gestione delle routine manuali all'avvio
TimerLcd=LcdCycle;   // azzera timer per routine init LCD
					      // ritardo di 150mS per far stabilizzare tutti i device all'avvio
LcdPutCharFlag=0;	   // azzera trasmissione byte su LCD
LcdStringPtr=0xFF;   // disabilita LcdStringPut

LcdVar[1][4]=0;             // menu 1 non si modifica
LcdVar[2][4]=0;             // menu 2 non si modifica
LcdVar[3][4]=0;             // menu 3 non si modifica
LcdVar[4][4]=1;             // menu 4 si modifica
LcdVar[5][4]=1;             // menu 5 si modifica

// non usate
LcdVar[4][2]=0; 
LcdVar[2][1]=0;
LcdVar[2][2]=0;
LcdVar[2][3]=0;               

// ---- init Keypad
FlagKbdIntr=0;		   // inizializza flag Keypad
KeypadInt=0;		   // inizializza flag Interrupt
FlagKbdStartRead=0;  // inizializza flag Lettura keypad
CursorStatus=0;      // posizione iniziale del cursore sui menu con modifica
FlagMenuMod=0;       // a default i menu non sono modificabili

// ---- init Bussola
TimerCmp=CmpCycle;   // azzera timer lettura bussola
FlagCmp=0;           // flag lettura bussola
FlagCmpCal=0;        // flag avvio procedura calibrazione bussola
FlagCmpReg15=0;      // flag calibrazione bussola

// ==== enable interrupt in questo momento parte il timer di sistema e quindi l'RTOS
INTCONbits.GIEH=1;	// abilitati tutti gli interrupt high
INTCONbits.GIEL=1;	// abilitati tutti gli interrupt low	


while (1)  // main loop
{	

//TimeKeeping------------------------------------
if (FlagTime)			// e' passato un altro millisecondo
{
	TimeKeeping();
}


//UserInterface----------------------------------
if (UserInterfaceFlag)  
/* flag di abilitazione di tutte le routine gestite manualmente: Keypad, LCD, EEPROM.
   Dal momento che sono eseguite molto raramente sono tutte ragruppate sotto un unico
   semaforo. Questo permette di risparmiare diversi controlli nel funzionamento normale.
*/
{
   UserInterface();     // (terminal.h)
}                        
                                            
   
//MotSpeed---------------------------------------
   if (PidTick)
   {
      MotSpeed();       // (movement.h)
      /* 
         Regola la velocita' dei motori
         questa routine e' eseguita solo su abilitazione della ISR, quest'ultima calcola
         la velocita' ogni x mSec e solo dopo averla calcolata ne permette la correzione.
         Il calcolo PID e' quindi eseguito piu' o meno con la stessa periodicita' 
      */
   }
      

//Space2Run--------------------------------------      
   if (Space2RunFlag) 
   {
      Space2Run();      // (movement.h)
      /* 
         Controlla lo spazio da percorrere
         qualche routine ha abilitato il controllo dello spazio percorso,
         ha impostato la velocita' (DesSpeedX), lo spazio da percorrere (Space2RunX)
         e la posizione di partenza (Space2RunXstart). 
         Questa routine sara' eseguita continuamente fino al raggiungimento della
         posizione desiderata, dopodiche' fermera' i motori e disabilitera' il flag
      */
   }
 
 
//Path-------------------------------------------
   if (Space2RunFlag == 0  && PathSeq[PathSeqPointer] != 0 && FlagStop == 0)
   {
      Path();          // (movement.h) 
      /* 
         Esegue, in sequenza, le manovre impostate
         se Space2RunFlag = 1, e' ancora in esecuzione una singola routine di movimento
         se PathSeq[PathSeqPointer] = 0 -> fine sequenza routines di movimento (EOL)
         se FlagStop = 1 -> la routine Stop ha fermato i motori
      
          La routine che deve usare Path
            -Imposta DesSpeedX = 0; // ferma motori
            -imposta PathSeq con passi da fare in ordine (fine seq = 0)
            -imposta PathSeqPointer = 0; // inizializza sequenza passi
            -imposta Space2RunFlag = 0; // reset di qualsiasi routine di movimento

         Se deve avanzare a velocita' costante non c'e' bisogno di chiamare Path(),
         si imposta semplicemente la velocita'
      */
   }
   
   
//Bumpers----------------------------------------
   Bumpers();        // (sensor.h)
   /*
      Controllo collisioni
      questi sensori hanno priorita' massima su tutti gli altri
      e resettano qualsiasi manovra eventualmente in esecuzione
   */
   

//Stop-------------------------------------------
   if (FlagStop)  
   {
      Stop();        // (movement.h)
      /*
         se qualche routine ha abilitato lo stop i motori si fermano per X mSec
      */
   }
   
 
//ReadAD-----------------------------------------
   if ((TimerAD) >= 7)
   {      
      ReadAD();      // (sensor.h) 
      /*
         legge le porte analogiche e mette i valori letti nelle rispettive variabili.
         TimerAD viene incrementato ogni 1ms dalla ISR
         La routine di conversione viene  chiamata solo dopo 7ms, in attesa Tacq, 
         anche se per il Tempo di acquisizione bastano 10-20us con il PIC 16F877 o il
         PIC 18F452 e 2-3us con il PIC 18F4525, non servono piu' di 20 letture al
         secondo per ognuno dei sensori.
         Tutte le porte AD sono utilizzate:
         EyeRl, EyeLl, EyeRf, EyeLf, FrR, FrL, FrC, Gas
         Ogni 7ms legge una porta, quindi ritorna a leggere la stessa porta dopo:
         7x7=49ms ,  1000/49=20 letture al secondo 
         Se aumentano le porte A/D da leggere, diminuire il tempo di attesa.
         Il  Timer viene  resettato alla fine del ciclo di conversione dalla routine  stessa
      */
   }
   

//Eye--------------------------------------------  
   if (Space2RunFlag == 0  && PathSeq[PathSeqPointer] == 0 && FlagStop == 0 && FlagEye)
   {
      Eye();         // (sensor.h) 
      /* 
         Controllo sensori di prossimita' ad IR
         Questa routine ha priorita' bassa rispetto alle altre routine di movimento
         viene quindi eseguita solo se non ci sono manovre in atto, anche se innescate
         dalla routine stessa
         se Space2RunFlag = 1, e' ancora in esecuzione una singola routine di movimento
         se PathSeq[PathSeqPointer] = 0 -> fine sequenza routines di movimento (EOL)
         se FlagStop = 1 -> la routine Stop ha fermato i motori
         se FlagEye = 0 -> non c'e' ancora stata una nuova lettura dei sensori, e' inutile
         	eseguire questa routine, si risparmiano molti cicli macchina
      */
   }
   

//Light------------------------------------------
   if (FlagLight)
   {
	   if (!FlagTargetLight)
	   {	
		   if (TimerSensor <= 0)
		   {
			   FlagTargetLight = 1;	// riabilita i sensori dopo "obiettivo raggiunto"
			}
		}
		else
		{
      	Light();    // (sensor.h)
  			/* 
         Controllo sensori di luce
         se FlagLight = 0 -> non c'e' ancora stata una nuova lettura delle fotoresistenze, 
         e' inutile eseguire questa routine, si risparmiano molti cicli macchina.
         Se la routine e' stata disabilitata perche' raggiunto obiettivo di luce:
       	(FlagTargetLight=0) controlla se e' scaduto il timer di disabilitazione.
         Se e' scaduto, riabilita la routine e non controlla piu' il timer
    		*/
      }
   }
   

//Gas--------------------------------------------
   if (FlagGas)
   {
	   if (!FlagTargetGas)
	   {	
		   if (TimerSensor <= 0)
		   {
			   FlagTargetGas = 1;	// riabilita i sensori dopo "obiettivo raggiunto"
			}
		}
		else
		{
      	Gas();      // (sensor.h) 
  			/* 
         Controllo sensori di gas
         se FlagGas = 0 -> non c'e' ancora stata una nuova lettura dei TGS822, 
         e' inutile eseguire questa routine, si risparmiano molti cicli macchina.
         Se la routine e' stata disabilitata perche' raggiunto sorgente di gas:
       	(FlagTargetGas=0) controlla se e' scaduto il timer di disabilitazione.
         Se e' scaduto, riabilita la routine e non controlla piu' il timer
    		*/
      }
   }


//Sound------------------------------------------
   if (FlagSound && FlagTargetLight && !Buzzer)
   {/* disabilita sensore suono mentre e' fermo per rivelazione luce altrimenti
	 	 potrebbe capitare in un loop infinito tra delay di rivelazione luce e suono
	 	 in caso di echi vicino a sorgenti luminose.
	 	 Disabilitato anche mentre suona il buzzer per evitare che riveli il suo tono
	 	 come segnale a 4KHz
    */   
	   if (!FlagTargetSound)
	   {	
		   if (TimerSensor <= 0)
		   {
			   FlagTargetSound = 1;	// riabilita i sensori dopo "obiettivo raggiunto"
			}
		}
		else
		{
	     	Sound();    // (sensor.h)
  			/* 
         Controllo sensore di suono
         Viene abilitata dalla routine ReadAD con la stessa temporizzazione della
         routine Gas.
         Se la routine e' stata disabilitata perche' raggiunto obiettivo di suono:
       	(FlagTargetSound=0) controlla se e' scaduto il timer di disabilitazione.
         Se e' scaduto, riabilita la routine e non controlla piu' il timer
    		*/
      }
   }
   

//Beep-------------------------------------------   
   if ((BeepCount>0) && (TimerBeep >= BeepTime))
   {
      	Beep(); 
      	/* 
         	Suona il buzzer
     		*/
   }
   
   
//DeadCorner-------------------------------------
   if (TimerDeadCorner >= DeadCornerTime)
   {
		/*
		controlla se il bot e' entrato in un angolo morto, ad ogni manovra si incrementa
		un contatore. Se il contatore supera un valore prefissato in una precisa
		finestra di tempo, il bot fa retromarcia. 
		In ogni caso si resettano contatore e timer dopo il timeout
		*/
	   	if (DeadCorner >= MaxDeadCorner)
	   	{
		   	// inverte la marcia
       	   PathSeq [0] = 2;   // indietro n cm
            PathSeq [1] = 51;  // gira di 135 o -135  randomicamente
            PathSeq [2] = 255; // cammina avanti a velocita' costante
            PathSeq [3] = 0;   // fine sequenza 
            PathSeqPointer = 0;// inizializza sequenza
		   }
		   DeadCorner = 0; // reset contatore e timer
		   TimerDeadCorner = 0;
	}


//I2C--------------------------------------------
	if (I2cEventFlag)			
	/* la  porta SSP ha comunicato, tramite ISR, che e' terminata la precedente azione I2c
		LA DESCRIZIONE DELLE PROCEDURE I2c PROSEGUE IN DETTAGLIO SU "settings.h"
	*/
	{
	  I2cLowService();		// si esegue l'azione successiva (myi2c.h)
	  I2cEventFlag=0;			// pronti per il prossimo interrupt
	  	// exit
	}
	else 
	{
		if (I2cBusyFlag)		// I2cHighService ha avviato una sequenza di
									// comunicazione I2c gestita tramite  ISR
									// solo alla fine la I2cLowService liberera' il bus
		{	
		// exit
		}
	
		else 
		{
			if ((I2c[I2cDevPtr].Flag.Rx > 0) || (I2c[I2cDevPtr].Flag.Tx > 0))
			// si controlla un altro record del buffer per vedere se le routine a 
			// livello piu' alto hanno inserito byte da scambiare nel proprio buffer

			{	
				I2cHighService();		// Inizializza una nuova sequenza di comunicazione 
											// e impegna il bus  (myi2c.h)
				// exit
		   }
	
			else 
			{
				if(I2cDevPtr < (I2cMaxDev-1))	// ci sono ancora record da controllare nel buffer
			   {
					I2cDevPtr ++;			// al prossimo giro controllera' il record successivo
  					// exit	
				}

			   else 							// ha controllato tutti i record del buffer
				{ 
					I2cDevPtr = 0;			// al prossimo giro ricomincera' dal primo device
 	  			}
			}
   	}
   }	

//CmpRead---------------------------------------
   if ((TimerCmp) <= 0)    
   {// e' tempo di leggere la bussola  
      if (!I2c[CmpPtr].Flag.Tx && !I2c[CmpPtr].Flag.Rx)
      {// se i buffer TX e RX I2C sono liberi si puo' avviare una nuova lettura    
        CmpRead();           // (sensor.h)
      }
      /*
         avvia la procedura di lettura o calibrazione della bussola tamite I2C
      */
   }
	
}  // end while
} // main
/*====================================================================================*/



/* Funzioni ***************************************************************************/


/*WriteEeprom *************************************************************************
scrive nell'Eeprom interna "Data" all'indirizzo "Adr"
*/

void WriteEeprom(unsigned char Data, unsigned char Adr)
{
   EEADR=Adr;
   EEDATA=Data;
   EECON1bits.WREN=1;    // abilita scrittura
   EECON2=0x55;
   EECON2=0xAA;
   EECON1bits.WR=1;      // avvia scrittura
} // WriteEeprom
/**************************************************************************************/


/*ReadEeprom **************************************************************************
ritorna il dato scritto nell'Eeprom interna all'indirizzo Adr
*/

unsigned char ReadEeprom (unsigned char Adr)
{
   EEADR=Adr;
   EECON1bits.EEPGD=0; // Point to DATA memory
   EECON1bits.CFGS=0;  // Access program FLASH or Data EEPROM memory
   EECON1bits.RD=1;    // EEPROM Read
   return EEDATA;

} // ReadEeprom
/**************************************************************************************/


/*TimeKeeping *************************************************************************
gestione temporizzazioni e calcolo velocita'
*/

void TimeKeeping (void)
 {// e' passato un'altro milliSecondo
    TimerLcd --;     // timer fino a  30  Sec per visualizzazione display  
    TimerBumpers ++; // timer fino a  30  Sec per routine bumpers
    TimerStop --;    // timer fino a  30  Sec per routine Stop Motori
    TimerBeep ++;		// timer fino a  30  Sec per routine Beep
    TimerSensor --;	// timer fino a  30  Sec per disabilitazione sensori
    TimerAD ++;      // timer fino a 255 mSec per routine conversione AD
    TimerDeadCorner ++;//timer fino a 30  Sec per routine rivelazione angoli morti
  	 TimerKeypad ++;  // timer fino a 255 mSec per debounce tasti
  	 TimerCmp --;     // timer fino a  30  Sec per routine lettura bussola
  	 RandomBit = RandomBit ^ 1; // ogni mSec viene invertito, usato come variabile random

   if (DeltaT >= 64)   //ogni 64 mSec campiona la velocita'
   {
      // velocita' = spazio/Tempo. Dividendo millesimi di mm con millisecondi il risultato e' in mm/Sec
      SpeedR = DivIntS64((EncoderRcount - EncoderRcountPrev) * stepR); // divide per 64 (DeltaT)
      EncoderRcountPrev = EncoderRcount;   // reset contatore
      
      SpeedL = DivIntS64((EncoderLcount - EncoderLcountPrev) * stepL); // ripete per l'altra ruota 
      EncoderLcountPrev = EncoderLcount;
      
      DeltaT=0;
      PidTick = 1;   // eseguito il calcolo della velocita' pu essere eseguito il calcolo del PID
   }
   else
   {
      DeltaT++;
   }
   
   FlagTime=0;		// azzera flag in attesa del prossimo millisecondo
   
} // TimeKeeping
/**************************************************************************************/


/*SequenzaStart ************************************************************************
Esce dallo stato "init" e inizializza le sequenze di movimento
*/

void SequenzaStart (void)
{
	InitFlag = 0;			// esce dallo stato init
		
	// ---- init movimento
	FlagBumpers = 1;  // disattiva bumpers
	TimerBumpers = 0; // per 1Sec 
	Space2RunFlag = 0;// reset di qualsiasi routine di movimento
	// fermo per 1 sec poi parte a velocita' normale
	     PathSeq [0] = 21;  // fermo 1 sec
        PathSeq [1] = 255; // cammina avanti a velocita' costante
        PathSeq [2] = 0;   // fine sequenza 
        PathSeqPointer = 0;// inizializza sequenza     
	BeepCount = 0; 			// suona 1 beep, per debug=0, altrimenti = 2 *=*=*=*=*=*
	TimerLcd=LcdCycle;		// tempo tra le visualizzazioni del display
	LcdInitStatus=0;	      // azzera contatore stati init LCD
   LcdPutCharFlag=0;	      // azzera trasmissione byte su LCD
   LcdStringPtr=0xFF;      // disabilita LcdStringPut
   LcdStatus=0;            // parte dal primo passo del ciclo
	DisplayStatus=90;			// disabilita display

	LedRossoOFF;				// led spenti
	LedGialloOFF;
	LedVerdeOFF;   
	LcdPutChar (0x01, 0, 4);// clear display

} // SequenzaStart
/**************************************************************************************/


/*Beep ********************************************************************************
esegue un numero di beep uguali al contenuto della variabile BeepCount/2
la routine che deve far suonare il beep, carica nella variabile BeepCount il numero 
di beep da eseguire moltiplicato per 2.
Ogni BeepTime ms viene chiamata Beep che inverte Buzzer e decrementa BeepCount.
Il buzzer sara' quindi per BeepTime ms ON e per BeepTime ms OFF (duty cycle 50%)
*/

void Beep (void)
{
	TimerBeep = 0; // azzera il contatore per il prossimo ciclo
	BeepCount--;   // decrementa il contatore di semicicli eseguiti
	Buzzer = Buzzer ^ 1; // se e' accceso lo spegne e viceversa
	
} // Beep
/**************************************************************************************/



/*===========================================================================*/

// Low priority interrupt vector

#pragma code LowVector = 0x18
void InterruptVectorLow (void)
{
  _asm
    goto InterruptHandlerLow //jump to interrupt routine
  _endasm
}

//----------------------------------------------------------------------------
// Low priority interrupt routine

#pragma code
#pragma interruptlow InterruptHandlerLow

/*===========================================================================*

/*IntServiceRoutine********************************************************************/
void InterruptHandlerLow (void)

{  
	 if (!EncoderRint && !EncoderLint && !PIR1bits.SSPIF && !PIR2bits.BCLIF) 
	 //se non e' interrupt previsto c'e' un errore
    {
	    	while (1)		// blocca l'esecuzione del programma
	 		{
	 		MotEnable = 0;	// ferma i motori
	 		Buzzer=1;		// suona il beep
	 		LedRossoOFF;	
	 		LedGialloON;
	 		LedVerdeON;
	 		}
	 }
	
    if (EncoderLint) // impulso dall'encoder SX?
    {
	 /* encoder in quadratura, se al cambiamento di stato i due segnali non sono in fase
       la direzione e' = FWD, altrimenti e' REW
    */     
      if (EncoderLpulse != EncoderLdir)
      {
          EncoderLcount ++;   // allora incrementa contatore
      }
      else
      {   
          EncoderLcount --;   // altrimenti decrementa contatore
      }
      EncoderLint = 0;
    }

	
    if (EncoderRint) // ripete per l'altro encoder
    {
      if (EncoderRpulse != EncoderRdir)
      {
          EncoderRcount ++;   // allora incrementa contatore
      }
      else
      {   
          EncoderRcount --;   // altrimenti decrementa contatore
      }
      EncoderRint = 0;   // reset interrupt flag
    }
    
    
    if (PIR1bits.SSPIF)					// un evento I2c e' stato completato
    {
	    	PIR1bits.SSPIF = 0;			// reset dell'interrupt I2c
	    	I2cEventFlag = 1;				// sara' eseguita la I2cLowService
	 }
	 
	  
	 if (PIR2bits.BCLIF)					// c'e' stata una collisione sul bus I2c
    {
	    	PIR2bits.BCLIF = 0;			// reset dell'interrupt I2c
	    	I2cEventFlag = 1;				// sara' eseguita la I2cLowService
	    	I2cBusCollFlag = 1;			// avverte della collisione
	 }
	     	       
}   // Low Priority IntServiceRoutine
/**************************************************************************************/


/*===========================================================================*/
// High priority interrupt vector

#pragma code HighVector = 0x08
void
InterruptVectorHigh (void)
{
  _asm
    goto InterruptHandlerHigh //jump to interrupt routine
  _endasm
}

//----------------------------------------------------------------------------
// High priority interrupt routine

#pragma code
#pragma interrupt InterruptHandlerHigh

/*===========================================================================*/


/*IntServiceRoutine********************************************************************/
void InterruptHandlerHigh (void)

{      
 if (INTCONbits.TMR0IF)   // l'interrupt e' stato generato dall'overflow del timer 0 ?
 {
    TMR0L -= 156;// ricarica il timer
    /*interrupt ogni 1mSec 
     periodo = CLKOUT * prescaler * TMR0
     1mSec ~= 1/(40.000.000/4) * 64 * 156
	 */
    
    FlagTime=1;	      // abilita aggiornamento temporizzazioni
      
    INTCONbits.TMR0IF = 0;// reset flag di interrupt
 }
 
 
 if  (KeypadInt)		   // e' stato premuto un tasto
  {
      FlagKbdIntr= 1;   // abilita la routine Keypad
      UserInterfaceFlag=1;// abilita la routine user interface
   	KeypadInt = 0;		// reset dell'interrupt 
  }
 		
}   // High Priority IntServiceRoutine
/**************************************************************************************/


/* Fine funzioni **********************************************************************/
