/*
contiene tutte le routine per la gestione del movimento
*/


/*Turn ********************************************************************************
routine per girare a destra o a sinistra di un angolo desiderato
la rotazione e' intorno al centro del bot senza avanzamento quindi le ruote girano una 
in senso opposto all'altra, alla stessa velocita'. 
Qui calcola e imposta i valori per girare dell'angolo desiderato.
L'interasse del bot e' di 140mm, la circonferenza lungo la quale girano le ruote e' quindi
di PI * 140 = 439,823mm. Per ruotare di 180 ogni ruota deve fare meta' circonferenza:
219.911micron (millesimi di millimetro), dividendo per 180 sappiamo che un grado
corrisponde a 1.222micron di spostamento di ogni ruota ((D * PI)/2/180).
Si considera:
0 la direzione di avanzamento 
da +1 a +180 la rotazione in senso orario
da -1 a -180 la rotazione in senso antiorario

operazioni:
imposta: Space2RunXstart
calcola e imposta: Space2RunX
imposta  velocita': DesSpeedX
imposta: Space2RunFlag a 1
*/

void Turn(long Gradi)
{
// fixed point base 2, la costante e' moltiplicata per 128, il risultato finale e' poi
// shiftato di 7 bit (diviso 128) per riportare il valore all'intero.

// Arco di circonferenza corrispondente ad un grado, in impulsi 
#define unGradoR  1220/stepR*128 // ruota destra
#define unGradoL  1220/stepL*128 // ruota sinistra

// #define corrGradi  3  
// se e' presente un errore sistematico assoluto, che quindi influisce in una percentuale 
// maggiore sui valori di angolo piccoli, inserire qui il parametro di correzione

Space2RunRstart = EncoderRcount; // punto di partenza
Space2RunLstart = EncoderLcount;

if (Gradi > 0)
{
   #ifdef corrGradi
   if (Gradi > 20) Gradi = Gradi - corrGradi; // non  corregge per angoli molto piccoli
   #endif
   
   DesSpeedR = manSpeedRew; // rotazione oraria, motore destro gira indietro
   DesSpeedL = manSpeedFwd; // ruota sinistra gira in avanti
  
   Space2RunR = DivLongS128((unGradoR * Gradi));// spazio (in impulsi encoder) da percorrere
   Space2RunL = DivLongS128((unGradoL * Gradi));   
}
else
{
   #ifdef corrGradi
   if (Gradi < -20) Gradi = Gradi + corrGradi;// non  corregge per angoli molto piccoli
   #endif
   
   DesSpeedR = manSpeedFwd; // rotazione antioraria, motore destro gira avanti
   DesSpeedL = manSpeedRew; // ruota sinistra gira indietro
   
   Space2RunR = DivLongS128((unGradoR * Gradi));// spazio (in impulsi encoder) da percorrere
   Space2RunL = DivLongS128((unGradoL * Gradi)); 
}
Space2RunFlag = 1; // inizia il controllo dello spazio da percorrere
} // Turn
/**************************************************************************************/

/*Walk ********************************************************************************
Calcola e inizializza le variabili per Space2Run, per camminare dello spazio desiderato
operazioni:
imposta: Space2RunXstart
Space2RunX e' impostata dalla routine chiamante
imposta  velocita': DesSpeedX
imposta: Space2RunFlag a 1
*/

void Walk(long Space)
{

Space2RunRstart = EncoderRcount; // punto di partenza
Space2RunLstart = EncoderLcount;

Space2RunR = Space / stepR; // da micron a conteggio encoder
Space2RunL = Space / stepL;

if (Space > 0) // avanti
{
   DesSpeedR = manSpeedFwd; // durante le manovre va piu' piano
   DesSpeedL = manSpeedFwd; 
}
else // indietro
{
   DesSpeedR = manSpeedRew; // durante le manovre va piu' piano
   DesSpeedL = manSpeedRew; 
}

Space2RunFlag = 1; // inizia il controllo dello spazio da percorrere

} // Walk
/**************************************************************************************/

/*Space2Run ***************************************************************************
controlla continuamente lo spazio percorso per pilotare i motori fino al raggiungimento
della posizione desiderata

operazioni:
controlla se spazio percorso >= spazio desiderato:
if yes, ferma motore corrispondente
se entrambi motori fermi: Space2RunFlag a 0 (si autodisabilita)

*/

void Space2Run (void)
{ 
/* qualche routine ha abilitato il controllo dello spazio percorso,
   ha impostato la velocita' (DesSpeedX), lo spazio da percorrere (Space2RunX)
   e la posizione di partenza (Space2RunXstart). 
   Questa routine sara' eseguita continuamente fino al raggiungimento della
   posizione desiderata, dopodiche' fermera' i motori e disabilitera' il flag
*/
    if (abs(EncoderRcount - Space2RunRstart) >= abs(Space2RunR))
    {
      DesSpeedR = 0;
    }
    
    if (abs(EncoderLcount - Space2RunLstart) >= abs(Space2RunL))
    {
      DesSpeedL = 0;
    }
    
    if ((DesSpeedR == 0) && (DesSpeedL == 0))
    {
      // raggiunto il punto desiderato si ferma per 100 mSec
      TimerStop = 100; // carica il contatore
      FlagStop = 1;  // abilita routine Stop
      Space2RunFlag = 0; // si autodisabilita
    }
} // Space2Run
/**************************************************************************************/

/*Stop ********************************************************************************
   Ferma i motori per X millisecondi (fino a 30 Sec)
*/
void Stop(void)
{
DesSpeedR = 0;   // Motori fermi
DesSpeedL = 0;
                   
if (TimerStop <= 0) //  il clock decrementa il contatore, quindi il valore caricato 
                    //  inizialmente in TimerStop e' = alla pausa in mSec 
   {
      LedVerdeOFF;   // spegne i led accesi al raggiungimento di un obietivo
      LedRossoOFF;	 // i led rimano accesi mentre i motori sono fermi
      LedGialloOFF;                  

      FlagStop = 0; // disabilita la routine Stop, riabilita la routine Path
   }
                    
} // Stop
/**************************************************************************************/

/*Path ********************************************************************************
   viene chiamata solo se (Space2RunFlag==0  & PathSeq[PathSeqPointer] != 0)
   se Space2RunFlag = 1, sono ancora in esecuzione routine movimento e quindi non pu
   passare allo step successivo
   se PathSeq[PathSeqPointer]= 0 la sequenza e' terminata e rilascia il controllo
   
   routine che deve usare Path
   -Imposta DesSpeedX = 0; // ferma motori
   -imposta PathSeq con passi da fare in ordine (fine seq = 0)
   -imposta PathSeqPointer = 0; // inizializza sequenza passi
   -imposta Space2RunFlag = 0; // reset di qualsiasi routine di movimento

   Se deve avanzare a velocita' costante si imposta semplicemente la velocita'

*/

 void Path (void)
{
      switch (PathSeq[PathSeqPointer])
      {   
         case 1:
                    Walk(1000000); // 1.000.000 micron = 1 metro avanti
                    break;
         case 2:
                    Walk(-30000); // -30.000 micron = 3 cm indietro
                    break;
         case 3:
                    Walk(-100); // -100 micron indietro,per azzerare i contatori
                    break;
         case 8:
                    Turn(10); // gira n gradi a Dx
                    DeadCorner++; // contatore per routine angolo morto
                    break;
         case 9:
                    Turn(-10);// gira n gradi a Sx
                    DeadCorner++; // contatore per routine angolo morto
                    break;
         case 12:
                    Turn(45);
                    DeadCorner++; // contatore per routine angolo morto
                    break;
         case 13:
                    Turn(-45);
                    DeadCorner++; // contatore per routine angolo morto
                    break;
         case 14:
                    Turn(90);
                    DeadCorner++; // contatore per routine angolo morto
                    break;
         case 15:
                    Turn(-90);
                    DeadCorner++; // contatore per routine angolo morto
                    break;
         case 18:
                    Turn(180);
                    break;
         case 19:
                    TimerStop = 100; // pausa per  100 mSec
                    FlagStop = 1;  // abilita routine Stop motori
                    break;
         case 20:
                    TimerStop = 10000; // pausa per  10 Sec
                    FlagStop = 1;  // abilita routine Stop motori
                    break;
         case 21:
                    TimerStop = 1000; // pausa per  1 Sec
                    FlagStop = 1;  // abilita routine Stop motori
                    break;
                    
         case 22:
                    TimerStop = 5000; // pausa per  5 Sec
                    FlagStop = 1;  // abilita routine Stop motori
                    break;
                    
         case 23:   // usato solo per la calibrazione della bussola
                    FlagCmpReg15=1; // calibra per ogni punto cardinale
                    TimerStop = 5000; // pausa per  5 Sec
                    FlagStop = 1;  // abilita routine Stop motori
                    break;
                                        
         case 50:
                    // gira in un verso o nell'altro pseudo-randomicamente
                    // diminuisce la probabilita' di ripetere sempre gli stessi percorsi
                    if (RandomBit)
                    	{
	                  	Turn(90);
	                 	}
	                 else
	                 	{
		                 	Turn(-90);
		               }
                    break;  
         case 51:
         			  // gira in un verso o nell'altro pseudo-randomicamente
                    // diminuisce la probabilita' di ripetere sempre gli stessi percorsi
                    if (RandomBit)
                    	{
	                  	Turn(135);
	                 	}
	                 else
	                 	{
		                 	Turn(-135);
		               }
                    break;       
         case 100:
                    DesSpeedR=0; // velocita' 0 cm/s
                    DesSpeedL=0; // sfrutta la funzione "brake" del metodo LAP
                    break;
     	   case 200:
                    DesSpeedR=lowSpeed; // velocita' ridotta
                    DesSpeedL=lowSpeed;
                    break;
         case 254:
                    FlagTargetLight = 1;	// riabilita routine sensori luce fino a fine manovra 
                    break;

         default:
                    DesSpeedR=constSpeed; // a default cammina a velocita' costante
                    DesSpeedL=constSpeed; // a default cammina a velocita' costante
                    break;
       }
      PathSeqPointer++;  // predispone al prossimo passo
} // Path
/**************************************************************************************/

/*MotSpeed ****************************************************************************
Controllo velocita' motori PID (Proportional + Integral + Derivative)
il ponte ad H e' usato in modalita' "Locked Anti Phase" (LAP):
CCPR2L collegato al motore R
CCPR1L collegato al motore L
*/
void MotSpeed (void)
{
/* 
   PWM = 256 -> motore fermo
   PWM = 512 -> velocita' massima FWD (circa  690 mm/sec)
   PWM =   0 -> velocita' massima REW (circa -690 mm/sec)
per riportare i conti effettuati sulla velocita' in valori di PWM:
const divisore =((690-0)/(512-256)); rapporto tra range velocita' e range PWM = 2,70.
Per mantenere la precisione di due decimali nei calcoli intermedi senza usare la virgola mobile
la costante e' moltiplicata per 256, il risultato finale e' poi diviso per 256 (fixed point base 2)
Per velocizzare ancora di piu' i calcoli, invece che fare PWM / divisore
e' meglio fare PWM * (1/divisore) -> la moltiplicazione e' piu' veloce della divisione.
Quindi: 1/2,70*256 = 94,81 approssimato a 95
*/
#define divisore  95 

/*
Per ottenere 19KHz di PWM con 40MHz di clock il valore del prescaler e' 4 e il valore caricato
in PR2 e' 127. In questo modo si possono usare solamente 9 bit dei 10 disponibili.
Il valore di PWM al 100% corrisponde quindi con 512 e il 50% (motori fermi) a 256.
*/

#define biasFwd  256 // (MinPWMFwd-(MinSpeed/divisore)), in pratica sono lo 0 del PWM FWD e REW, in questo caso 
#define biasRew  256 // (MinPWMRew+(Minspeed/divisore)), non c'e' dead-band, la regolazione e' continua da 0 a 1023 

#define intUpperLimit  10230  //limite superiore della correzione integrale
#define intLowerLimit  -10230 //limite inferiore della correzione integrale
#define devUpperLimit  10230  //limite superiore della correzione derivativa
#define devLowerLimit  -10230 //limite inferiore della correzione derivativa
#define speedUpperLimit  689 //limite superiore velocita'
#define speedLowerLimit -689 //limite inferiore velocita'

long PWMR;   // risultato del calcolo del PID, ancora e' in mm/sec
long PWML;

int ErroreR = 0;   //errore attuale
int ErroreL = 0;

int DevR   = 0;   //componente derivativa attuale
int DevL   = 0;   //componente derivativa attuale

/* questa routine e' eseguita solo su abilitazione della ISR, quest'ultima calcola la velocita' ogni x mSec 
   e solo dopo averla calcolata ne permette la correzione
   il calcolo PID e' quindi eseguito piu' o meno con la stessa periodicita' 
*/

   PidTick = 0;   // resetta il flag per il prossimo giro
   ErroreR = (DesSpeedR - SpeedR);   //errore attuale = velocita' desiderata - velocita' attuale
   ErroreL = (DesSpeedL - SpeedL);   //

   /* Fattore I
      esegue la sommatoria algebrica dell'errore, cioe' lo integra. 
      L'integrale dell'errore di velocita' e' lo spazio perso, o guadagnato, rispetto a quello voluto.
      In questo modo anche se una ruota rallenta a causa di un ostacolo perdendo terreno rispetto all'altra
      questo viene riguadagnato (entro certi limiti)
   */
   IntR =IntR + (ErroreR * ki);
   if (IntR > intUpperLimit)   // la quantita' assoluta di correzione e' limitata per non saturare i contatori
   {
      IntR = intUpperLimit;
   }
   if (IntR < intLowerLimit)
   {
      IntR = intLowerLimit;
   }

   IntL = IntL + (ErroreL * ki);
   if (IntL > intUpperLimit)
   {
      IntL = intUpperLimit;
   }
   if (IntL < intLowerLimit)
   {
      IntL = intLowerLimit;
   }

   /* 
   Fattore D
   calcola la variazione dell'errore rispetto alla misura precedente, 
   essendo l'intervallo di tempo di misura costante, corrisponde con la derivata dell'errore di velocita'
   e quindi con l'accellerazione
   */
   DevR = kd * (ErroreR - ErroreRprev);
   ErroreRprev = ErroreR;
   if (DevR > devUpperLimit)   // la quantita' assoluta di correzione e' limitata per non saturare i contatori
   {
      DevR = devUpperLimit;
   }
   if (DevR < devLowerLimit)
   {
      DevR = devLowerLimit;
   }

   DevL = kd * (ErroreL - ErroreLprev);
   ErroreLprev = ErroreL;
   if (DevL > devUpperLimit)
   {
      DevL = devUpperLimit;
   }
   if (DevL < devLowerLimit)
   {
      DevL = devLowerLimit;
   }

   /* Il risultato e' diviso 16 per sfruttare il decimale senza usare variabili float, 
      le costanti erano infatti moltiplicate per 16, in questo modo si usa la precisione 
      del decimale nei calcoli intermedi pur avendo un intero come risultato finale
   */
   
   PWMR = DivLongS16((long)(ErroreR * kp + IntR + DevR));   // P + I + D
   PWML = DivLongS16((long)(ErroreL * kp + IntL + DevL));   //   

   /* converte la velocita' da mm/sec (signed int moltiplicato 256) a valore da impostare nei registri CCPRxL (char)
      PWM da -690 a 690 -> CCPRxL:CCPxCON<5:4> da 0 a 512
   */
   if (PWMR > 0)   // Motore Destro in avanti
   {
      if (PWMR > speedUpperLimit)   // anche qui si limita per non sforare le variabili
      {
         PWMR = speedUpperLimit;
      }
      PWMR = DivLongS256(PWMR * divisore)  + biasFwd; //  diviso 256 per riportare alla giusta dimensione
      SetDCPWM2(PWMR);
   }
   else   // indietro
   {
      if (PWMR < speedLowerLimit)
      {
         PWMR = speedLowerLimit;
      }
      PWMR = biasRew + DivLongS256(PWMR * divisore); 
      SetDCPWM2(PWMR);
   }

   if (PWML > 0)   // Motore Sinistro in avanti
   {
      if (PWML > speedUpperLimit)
      {
         PWML = speedUpperLimit;
      }
      PWML = DivLongS256(PWML * divisore) + biasFwd;
      SetDCPWM1(PWML);
   }
   else   // indietro
   {
      if (PWML < speedLowerLimit)
      {
         PWML = speedLowerLimit;
      }
      PWML = biasRew + DivLongS256(PWML * divisore);
      SetDCPWM1(PWML);   
   }

   
} //  MotSpeed
/**************************************************************************************/
